#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include "cygwin.h"

#include "misc.h" // only for timer
#include "cSocket.h"
#include <map>
#include <list>
#include <algorithm>
#include <errno.h>

#define TCP_ACK 40

// Values important for the UL/DL Management
size_t	global_ul_count	=  0;
size_t 	global_dl_count	=  0;
int 	ul_left		=  0;
 int	dl_left		=  0;
size_t 	ulSecond	= 10 * 1024;
size_t 	dlSecond	= 80 * 1024;
struct in_addr		 my_ip;

const char *txtType (const cSocket *list1) {{{
	if (list1 == NULL) return "(NULL)";
	switch(list1->Type) {
	case tNone    : return "none    ";
	case tSocket  : return "socket  ";
	case tServer  : return "server  ";
	case tSource  : return "source  ";
	case tGui     : return "gui     ";
	case tCommand : return "source  ";
	case tKademlia: return "kademlia";
		}
	return "";
	}}}

const char* GetSockTypeString( int type ) {{{
	switch( type ) {
		case SOCK_STREAM:	return "SOCK_STREAM";	/* stream socket */
		case SOCK_DGRAM:	return "SOCK_DGRAM";	/* datagram socket */
		case SOCK_RAW:		return "SOCK_RAW";		/* raw-protocol interface */
		case SOCK_RDM:		return "SOCK_RDM";		/* reliably-delivered message */
		case SOCK_SEQPACKET:return "SOCK_SEQPACKET";/* sequenced packet stream */
		default: return "UNKNOWN SOCK TYPE!";
		}
	}}}	

void cSocket::resetAccept(void) {{{
	log (1, "resetAccept: Closing Socket!!! fd_new=%u\n", fd_new );
	closesocket(fd_new);
	fd_new =INVALID_SOCKET;
	}}}

typedef std::map<SOCKET,cSocket*> cSocketMap;
cSocketMap g_SocketMap;
cSocketMap::iterator g_CurSocket = g_SocketMap.end();
static void  unregisterSocket (SOCKET fd) {{{
	// we must protect g_CurSocket so that it does not become invalid due to removal
	if( g_CurSocket != g_SocketMap.end() && (*g_CurSocket).first == fd )
		++g_CurSocket;
	g_SocketMap.erase(fd);
	}}}
static void    registerSocket (SOCKET fd, class cSocket *addr) {{{
	ASSERT(addr->fd != 0);
	g_SocketMap[fd] = addr;
	}}}
class cSocket  *retriveSocket (int fd) {{{
//	Start_Timer(50);
	cSocket *pRet = NULL;
	cSocketMap::iterator i;
	// find cSocket* corresponding to given fd
	i = g_SocketMap.find(fd);
	if( i != g_SocketMap.end() ) pRet = (*i).second; // it exists, so return it
	if (pRet != NULL) ASSERT(pRet->fd != 0);
//	Ende_Timer(50,"retrieveSocket (std::map)",0);
	return pRet;
	}}}
class cSocket *socket_First (void) {{{
	g_CurSocket = g_SocketMap.begin();
	if( g_CurSocket == g_SocketMap.end() ) return NULL;
	return (*g_CurSocket).second;
	}}}
class cSocket *socket_Next  (void) {{{
	if( g_CurSocket == g_SocketMap.end() ) return NULL;
	++g_CurSocket;
	if( g_CurSocket == g_SocketMap.end() ) return NULL;
	return (*g_CurSocket).second;
	}}}
// using namespace std;
cSocketList event_list;
cSocketList timer_list;

template <class T>          cDLL<T>:: cDLL (cDLL<T> *after, T *neu) {{{
	this->next = after->next;
	this->prev = after;
	this->prev->next = this;
	this->next->prev = this;
	this->data = neu;
	}}}
template <class T>          cDLL<T>::~cDLL () {{{
	prev->next = next;
	next->prev = prev;
	next = NULL;
	prev = NULL;
	data = NULL;
	}}}
template <class T> cDLL<T> *cDLL<T>::add_front(T *neu) {{{
	return new cDLL<T>(this, neu); 
	}}}
template <class T> cDLL<T> *cDLL<T>::add_back(T *neu) {{{
	return new cDLL<T>(this->prev, neu); 
	}}}
template <class T> void     cDLL<T>::remove (cDLL<T> **old) {{{
	if (*old == NULL) return;
	delete *old;
	(*old) = NULL;
	}}}

//#undef  HAVE_ASYNCIO
//#define HAVE_POLLIO	1

#ifdef _WIN32
#define GetErrno()              WSAGetLastError()
#define SOCKET_WOULD_BLOCK	WSAEWOULDBLOCK
#define SOCKET_RETRY		WSAEWOULDBLOCK
// {{{
u_long read_avail = 0;	// cSocket::do Read
HINSTANCE asyncIO_prc;
HWND 	  asyncIO_win;
ATOM 	  asyncIO;

void Win32_PrintErrorString( int err ) {{{
	char*  msg;
  	FormatMessage( 
    		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    		NULL,
    		err,
    		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
    		(LPTSTR) &msg,
    		0,
    		NULL 
  		);
  	printf( msg );
  	LocalFree( msg );
	}}}

void waitEvent(int ) {{{
	MSG msg;
	BOOL bRet;
	if ((bRet = GetMessage( &msg, NULL, 0, 0 )) != 0) {   // Main message loop: 
		if (bRet == -1) {
			printf( "GetMessage error: " );
			Win32_PrintErrorString( GetLastError() );
		}
		else {
			TranslateMessage(&msg); 
			DispatchMessage(&msg); 
//			cSocket::hdlEvent();
		}
	}
	}}}

char* GetWSAErrorString( int err ) {{{	// WINDOWS spezific
char* pMsg;
	switch(err) {
  	case WSAVERNOTSUPPORTED: pMsg = "WSAVERNOTSUPPORTED: version of WinSock not supported"; break;
  	case WSASYSNOTREADY:     pMsg = "WSASYSNOTREADY: WinSock not present or not responding"; break;
  	case WSAEINVAL:          pMsg = "WSAEINVAL: app version not supported by DLL"; break;
  	case WSAHOST_NOT_FOUND:  pMsg = "WSAHOST_NOT_FOUND: Authoritive: Host not found"; break;
  	case WSATRY_AGAIN:       pMsg = "WSATRY_AGAIN: Non-authoritive: host not found or server failure"; break;
  	case WSANO_RECOVERY:     pMsg = "WSANO_RECOVERY: Non-recoverable: refused or not implemented"; break;
  	case WSANO_DATA:         pMsg = "WSANO_DATA: Valid name, no data record for type"; break;
  	case WSANOTINITIALISED:  pMsg = "WSANOTINITIALISED: WSA Startup not initialized"; break;
  	case WSAENETDOWN:        pMsg = "WSAENETDOWN: Network subsystem failed"; break;
  	case WSAEINPROGRESS:     pMsg = "WSAEINPROGRESS: Blocking operation in progress"; break;
  	case WSAEINTR:           pMsg = "WSAEINTR: Blocking call cancelled"; break;
  	case WSAEAFNOSUPPORT:    pMsg = "WSAEAFNOSUPPORT: address family not supported"; break;
  	case WSAEMFILE:          pMsg = "WSAEMFILE: no file descriptors available"; break;
  	case WSAENOBUFS:         pMsg = "WSAENOBUFS: no buffer space available"; break;
  	case WSAEPROTONOSUPPORT: pMsg = "WSAEPROTONOSUPPORT: specified protocol not supported"; break;
  	case WSAEPROTOTYPE:      pMsg = "WSAEPROTOTYPE: protocol wrong type for this socket"; break;
  	case WSAESOCKTNOSUPPORT: pMsg = "WSAESOCKTNOSUPPORT: socket type not supported for address family"; break;
  	case WSAENOTSOCK:        pMsg = "WSAENOTSOCK: descriptor is not a socket"; break;
  	case WSAEWOULDBLOCK:     pMsg = "WSAEWOULDBLOCK: socket marked as non-blocking and SO_LINGER set not 0"; break;
  	case WSAEADDRINUSE:      pMsg = "WSAEADDRINUSE: address already in use"; break;
  	case WSAECONNABORTED:    pMsg = "WSAECONNABORTED: connection aborted"; break;
  	case WSAECONNRESET:      pMsg = "WSAECONNRESET: connection reset"; break;
  	case WSAENOTCONN:        pMsg = "WSAENOTCONN: not connected"; break;
  	case WSAETIMEDOUT:       pMsg = "WSAETIMEDOUT: connection timed out"; break;
  	case WSAECONNREFUSED:    pMsg = "WSAECONNREFUSED: connection refused"; break;
  	case WSAEHOSTDOWN:       pMsg = "WSAEHOSTDOWN: host down"; break;
  	case WSAEHOSTUNREACH:    pMsg = "WSAEHOSTUNREACH: host unreachable"; break;
  	case WSAEADDRNOTAVAIL:   pMsg = "WSAEADDRNOTAVAIL: address not available"; break;
  	default:                 pMsg = "WSA UNKNOWN ERROR!"; break;
	}
	return pMsg; 
	}}}
void Win32_SocketEventToString( WORD ev, std::string& str ) {
	if( ev & FD_ACCEPT )    str += "ACCEPT ";
	if( ev & FD_CONNECT )   str += "CONNECT ";
	if( ev & FD_OOB )       str += "OOB ";
	if( ev & FD_READ )      str += "READ ";
	if( ev & FD_WRITE )     str += "WRITE ";
	if( ev & FD_CLOSE )     str += "CLOSE ";
//	if( ev & FD_QOS )       str += "QOS ";
//	if( ev & FD_GROUP_QOS ) str += "GROUP_QOS ";
}
LRESULT CALLBACK handleSocket (HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {{{ // WINDOWS spezific
	if (uMsg != WM_USER) return DefWindowProc(hwnd, uMsg, wParam, lParam);	// not our async message
	int fd = (int)wParam;
	class cSocket *sock= retriveSocket (fd);
	int  event = 0;
	WORD win_event = LOWORD(lParam);
	int err = HIWORD(lParam);
	std::string strEvent;

	if( err || sock == NULL ) {
		Win32_SocketEventToString( win_event, strEvent );
		printf( "handleSocket: fd = %i, event = %s\n", fd, strEvent.c_str() );
		if( err ) printf( "  ERRNO:%i => %s\n", err, GetWSAErrorString(err) );
		if( sock == NULL ) printf( "  Received event for an unregistered socket.\n" );
		else sock->last_err = err;
		return 0;
		}

	if( win_event & FD_CLOSE )	{
		delete sock;
		return 0;
		}
	if( win_event & (FD_ACCEPT | FD_READ) )
		event |= 1;
	if( win_event & (FD_CONNECT | FD_OOB | FD_WRITE) )
		event |= 2;

	if( event )	sock->doWork( event );
	else {
		Win32_SocketEventToString( win_event, strEvent );
		printf( "handleSocket: unhandled event = %s\n", strEvent.c_str() );
		}

	return 0;
	}}}
// TimerProc is called every second.
VOID CALLBACK TimerProc( HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime ) {{{
	// *** call every_sec() here ***
//	printf("TimerProc: called every second. Time (ms) = %li\n", dwTime);
	}}}
void               initSocket (void) {{{	// WINDOWS spezific
	const char name[] = "asyncio";
	WNDCLASS wc;					// Hidden window for async-io message tracking
	wc.style=0;
	wc.lpfnWndProc=(WNDPROC)handleSocket;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hIcon=0;
	wc.hCursor=0;
	wc.hbrBackground=0;
	wc.lpszMenuName=(LPSTR)0;
	wc.lpszClassName= name;
	asyncIO_prc = (HINSTANCE) wc.hInstance = GetModuleHandle(NULL);
	if (!(asyncIO = RegisterClass(&wc))) {
		printf("RegisterClass failed :-(\n");
		exit(0);
		}
	if (!(asyncIO_win = CreateWindow(name,"win-io",WS_POPUP,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,0,0,asyncIO_prc,0))) {
		printf("CreateWindow failed :-(\n");
		exit(0);
		}
	WSADATA wsadata;
	if (0 == SetTimer(asyncIO_win, 1, 1000, TimerProc)) { // TimerProc is called once every second
		printf("Timer failed\n");
		exit(0);
		}
	if (WSAStartup(0x0101, &wsadata)) {
		printf("WSAStartup failed\n");
		WSACleanup();
		exit(0);
		}
	if (wsadata.wVersion < 0x0101) {
		printf("Wrong winsock version\n");
		WSACleanup();
		exit(0);
		}
	}}}
void               stopSocket (void) {{{ // WINDOWS spezific
	WSACleanup();
	DestroyWindow(asyncIO_win);
	UnregisterClass((LPCTSTR) MAKEINTATOM(asyncIO), asyncIO_prc);
	for (class cSocket *akt = socket_First (); akt!= NULL; akt= socket_Next ()) delete akt;
	}}}
void     cSocket::doUnblock   (void) {{{ // WINDOWS spezific
	if (fd == INVALID_SOCKET) return;
        u_long command = 1;
        int ret = ioctlsocket (fd, FIONBIO, &command);        //  Redirect events and set non-blocking mode
        if(ret != 0) {
		printf("Seting socket to nonblocking failed ERR: %i\n", ret);
                }
	if (WSAAsyncSelect(fd, asyncIO_win, WM_USER, FD_READ|FD_WRITE|FD_ACCEPT|FD_CLOSE|FD_OOB) < 0) {
		printf("ERROR: WSAAsyncSelect failed\n");
		exit(0);
		}
	}}}
void     cSocket::printError  (int err) {{{	// WINDOWS spezific
	printf("ERRNO %u => %s\n", err, GetWSAErrorString(err) );
	printf("  fd %i, port %u, peer %s:%u\n", fd, this_port, inet_ntoa(peer_ip), peer_port );
	}}}
// }}}
#else
#define GetErrno()              errno
#define SOCKET_WOULD_BLOCK  	EINPROGRESS
#define SOCKET_RETRY		EAGAIN
// {{{
int read_avail = 0;	// cSocket::do Read
#ifdef HAVE_ASYNCIO
#define USE_SIGNAL      (SIGRTMIN+5)
void waitEvent (int     ) {{{ // int usec
	timespec timeout;
	sigset_t 	block_rtmin;
	siginfo_t 	info;
	sigemptyset (&block_rtmin);
	sigaddset   (&block_rtmin, SIGTERM);
	sigaddset   (&block_rtmin, SIGINT);
	sigaddset   (&block_rtmin, USE_SIGNAL);
	sigaddset   (&block_rtmin, SIGALRM   );
	class cSocket *sock = NULL;
	while (running) {
		timeout.tv_sec  = 1;                            // Set timeout for the select
		timeout.tv_nsec = 0;
		int ret = sigtimedwait (&block_rtmin, &info, &timeout);
		if (!running) return;
		if (ret < 0) continue;
		if (info.si_signo == SIGINT ) { running = false; return; }
		if (info.si_signo == SIGTERM) { running = false; return; }
		if (info.si_signo == SIGALRM) { every_sec ();  continue; }
		if (info.si_signo == USE_SIGNAL) {
			sock = retriveSocket (info.si_fd);
			if (sock == NULL) return;
//			printf("EVENT %4i - %15s - %i\n", info.si_fd, inet_ntoa(sock->peer_ip), info.si_code);
			info.si_code |= sock->mask;
			sock->mask &= 32;
			if ( 4 == (info.si_code &  4)) sock->last_err = info.si_errno;
			if ( 0 == (info.si_code & 16)) sock->doWork(info.si_code); // ORDER IS IMPORTANT
			if (16 == (info.si_code & 16)) delete sock;
			continue;
			}
		printf("SIGNAL %i unhandeld\n", info.si_signo);
		} // while (running)
	}}}
void handleSocket (int, siginfo_t *, void *) { printf("handleSocket\n"); }
#elif HAVE_POLLIO
#define POLL_LIMIT 8192
void waitEvent (int usec) {{{
	struct pollfd poll_fds[POLL_LIMIT]; // limited file descriptors!
	class cSocket *akt = NULL;
	unsigned int i = 0;
	for (akt = socket_First (); akt != NULL; akt = socket_Next ()) {
		poll_fds[i].fd = akt->fd;
		poll_fds[i].events = POLLIN | POLLERR | POLLHUP | POLLNVAL;
		if (akt->send_len > 0)   poll_fds[i].events |= POLLOUT;
		if (akt->connected == 0) poll_fds[i].events |= POLLOUT;
		poll_fds[i].revents = 0;
		if (++i >= POLL_LIMIT) break;
		}
	int count = poll (poll_fds, i, (usec / 1000));
	if (count <= 0) {
		if (count < 0) ; // printf("Error %i in poll()\n", errno);
		return;
		}
	i = 0;
	for (akt = socket_First (); akt != NULL; akt = socket_Next ()) {
		int mask = akt->mask;
		akt->mask &= 32;
		if ((poll_fds[i].revents & POLLIN  ) != 0) mask |=  1;
		if ((poll_fds[i].revents & POLLPRI ) != 0) { printf("urgent data\n"); mask |= 16;  }
		if ((poll_fds[i].revents & POLLOUT ) != 0) mask |=  2;
		if ((poll_fds[i].revents & POLLERR ) != 0) mask |=  4;
		if ((poll_fds[i].revents & POLLHUP ) != 0) mask |= 16;
		if ((poll_fds[i].revents & POLLNVAL) != 0) mask |= 16;
		if ((mask & 16) == 0) akt->doWork(mask);
		mask |= akt->mask;
		if ((mask & 16) != 0) delete akt;
		if (++i >= POLL_LIMIT) break;
		}
	}}}
# else
void waitEvent (int usec) {{{
	struct timeval timeout;
	int count;
	timeout.tv_sec  = (usec / 1000000); // Set timeout for the select()
	timeout.tv_usec = (usec % 1000000); 
	fd_set	fd_1;
	fd_set	fd_2;
	fd_set	fd_4;
	FD_ZERO(&fd_1);
	FD_ZERO(&fd_2);
	FD_ZERO(&fd_4);
	int max = 0;
	class cSocket *akt = NULL;
	for (akt = socket_First (); akt != NULL; akt = socket_Next ()) {
		if (akt->fd < 0) continue;
		if (akt->fd > 1023) continue;
		if (akt->fd > max) max = akt->fd + 1;
					 FD_SET(akt->fd, &fd_1);
		if (akt->send_len  >  0) FD_SET(akt->fd, &fd_2);
		if (akt->connected == 0) FD_SET(akt->fd, &fd_2);
				 	 FD_SET(akt->fd, &fd_4);
		}
	count = select (max + 1, &fd_1, &fd_2, &fd_4, &timeout);
	if (count <= 0) {
		if (count < 0) ; // printf("Error %i in select()\n", errno);
		return;
		}
	for (akt = socket_First (); akt != NULL; akt = socket_Next ()) {
		if (akt->fd <    0) continue;
		if (akt->fd > 1023) continue;
		if (count   <=   0) continue;
		if (FD_ISSET(akt->fd, &fd_1)) { akt->addEvent(1); count--; }
		if (FD_ISSET(akt->fd, &fd_2)) { akt->addEvent(2); count--; }
		if (FD_ISSET(akt->fd, &fd_4)) { akt->addEvent(4); count--; }
		}
	}}}
#endif
void               initSocket (void) {{{ // LINUX spezific
#ifdef HAVE_ASYNCIO	
struct sigaction act;
	bzero (&act, sizeof(act));
	act.sa_sigaction = handleSocket;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_SIGINFO;
	if (0 != sigaction(USE_SIGNAL, &act, NULL)) {
		printf("USE_SIGNAL => %i\n", USE_SIGNAL);
		printf("USE_SIGNAL => 32 would be correct\n");
		printf("sigaction => %i : %s\n", errno, strerror(errno));
		abort();
		}
	// Block "USE_SIGNAL" so we can dequeue it in the wait process
	sigset_t block_rtmin;
	sigemptyset (&block_rtmin);
	sigaddset   (&block_rtmin, USE_SIGNAL);
	sigaddset   (&block_rtmin, SIGALRM   );
	sigprocmask (SIG_BLOCK, &block_rtmin, NULL);
	struct itimerval cur_timer;
	struct itimerval old_timer;
	cur_timer.it_interval.tv_sec  = 1;
	cur_timer.it_interval.tv_usec = 0;
	cur_timer.it_value   .tv_sec  = 1;
	cur_timer.it_value   .tv_usec = 0;
	setitimer(ITIMER_REAL, &cur_timer, &old_timer);
#endif	
	}}}
void               stopSocket (void) {{{ // LINUX spezific
	for (class cSocket *akt = socket_First (); akt!= NULL; akt= socket_Next ()) delete akt;
	}}}
void     cSocket::doUnblock   (void) {{{ // LINUX spezific
#ifdef HAVE_ASYNCIO
	int async = O_ASYNC | O_NONBLOCK;
#else	
	int async =           O_NONBLOCK;
#endif	

	if (fd == INVALID_SOCKET) return;
	int opt = 1; // sec to wait for data
	if (0 != fcntl (fd, F_SETFL , async | fcntl (fd, F_GETFL, 0))) {{{ // abort
		printf("fcntl (fd, F_SETFL , O_ASYNC) => %i : %s\n", errno, strerror(errno));
		abort();
		}}}
	if (type != SOCK_DGRAM)
		if (0 != setsockopt(fd, SOL_TCP, TCP_NODELAY, &opt, sizeof(int))) {
			printf("setsockopt(fd, SOL_TCP, TCP_NODELAY, [1], 4)  => %i : %s\n", errno, strerror(errno));
			}
#ifdef HAVE_ASYNCIO	
	if (0 != fcntl (fd, F_SETOWN, getpid())) {{{ // abort
		printf("fcntl (fd, F_SETOWN, getpid()) => %i : %s\n", errno, strerror(errno));
		abort();
		}}}
	if (0 != fcntl (fd, F_SETSIG, USE_SIGNAL)) {{{ // abort
		printf("fcntl (fd, F_SETSIG, USE_SIGNAL) => %i : %s\n", errno, strerror(errno));
		abort();
		}}}
	struct pollfd check;
	check.fd = fd;
	check.events = POLLIN | POLLOUT | POLLERR;
	if (0 < poll (&check, 1, 0)) {
		mask &= 32;
//		printf("POLL 1:%i 2:%i 4:%i\n", check.revents & POLLIN ,check.revents & POLLOUT,check.revents & POLLERR);
		if (check.revents & POLLIN ) mask |= 1;
		if (check.revents & POLLOUT) mask |= 2;
		if (check.revents & POLLERR) mask |= 4;
//		if (mask != 0) doWork(mask);
		}
#endif		
	}}}
void     cSocket::printError  (int err) {{{	// LINUX spezific
	ASSERT(err != 11);
	ASSERT(err != 13);
//	ASSERT(err != 22);	// EINVAL
	ASSERT(err != SOCKET_WOULD_BLOCK);
	printf("%15s:%5u ERRNO %5u => %s\n", inet_ntoa(peer_ip), peer_port, err , strerror(err));
	fflush(stdout);
	}}}
// }}}
#endif
void     cSocket::Init0     (void) {{{
	Type 	  = tNone;
	event 	  = NULL;
	mask 	  = 0;
	prio 	  = 1;
	fd_new 	  = INVALID_SOCKET;
	timer 	  = NULL;
	event 	  = NULL;
	send_buf  = NULL;
	send_len  = 0;
	read_buf  = NULL;
	read_len  = 0;
	last_read = 0;
	last_send = 0;
	since 	  = currentTime;
	work_last = currentTime;
	timer 	  = NULL;
	}}}
void     cSocket::Init      (void) {{{
	if (fd != INVALID_SOCKET) {
		doUnblock();
		registerSocket (fd, this);
		}
	timer = timer_list.add_back (this);
	}}}
cSocket         :: cSocket  (class cSocket *FD) {{{
	Init0();
	fd = FD->fd_new;
	if (fd == INVALID_SOCKET) {
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		addEvent(16);
		return;
		}
	last_err = 0;
	FD->fd_new = INVALID_SOCKET;
	type = SOCK_STREAM;
	Init();
	connected = 1;
	ready_send = 1;
	dl_count = ul_count = 0 ;
	struct sockaddr_in addr;
	socklen_t len = sizeof(addr);
	getpeername(fd, (struct sockaddr*)&addr, &len);
	this_ip.s_addr   = 0;
	this_port = FD->this_port;
	peer_port = 0;
	peer_ip   = addr.sin_addr;
	peer_port = addr.sin_port;
	last_conn = currentTime;
//	printf("NEW_FD(0) %4i\n", fd);
	}}}
cSocket         :: cSocket  (unsigned int IP, unsigned short PORT, int TYPE) {{{
	Init0();
	int err = 0;
	last_err = 0;
	type = TYPE;
	if (PORT == 0 || (type != SOCK_STREAM && type != SOCK_DGRAM)) {
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		addEvent(16);
		return;
		}
	fd = socket (AF_INET, type, 0);
	if (fd == INVALID_SOCKET) {{{
		printError (GetErrno());
		addEvent(16);
		return;
		}}}
//	printf("NEW_FD(1) %4i\n", fd);
	struct sockaddr_in srv_addr;
    	srv_addr.sin_family      = AF_INET;
    	if ( strcmp(bind_ip,"0") == 0 ) srv_addr.sin_addr.s_addr   = INADDR_ANY;
    	srv_addr.sin_addr.s_addr     = inet_addr(bind_ip);
    	srv_addr.sin_port        = htons(0);
    	if (0 != bind(fd, (struct sockaddr*)&srv_addr, sizeof(srv_addr))) {{{
		err = GetErrno();
	        printf("Bind to socket failed %s\n", strerror(errno));
		printError (err);
		addEvent(16);
        	return;
    		}}}
	Init();
	dl_count = ul_count = 0 ;
	this_ip.s_addr	  = 0;
	this_port = 0;
	peer_port = PORT;
	peer_ip.s_addr = IP;
	struct sockaddr_in server_addr;
	memset(&server_addr, 0, sizeof(struct sockaddr_in));
	server_addr.sin_family  	= AF_INET;
	server_addr.sin_addr.s_addr	= IP;
	server_addr.sin_port    	= htons (PORT);
	size_t ret = connect(fd, (struct sockaddr *)&server_addr, sizeof(server_addr));
	err = GetErrno();
	if (ret == (size_t)INVALID_SOCKET && err != SOCKET_WOULD_BLOCK) {
		printf ("BIND ");
		printError (err);
		addEvent(16);
		return;
		}
	connected = (ret == 0);
	ready_send = (ret == 0);
	}}}
cSocket         :: cSocket  (unsigned short PORT, int TYPE) {{{
	Init0();
	if (TYPE == SOCK_STREAM) connected = 2;
	else			 connected = 3;
	dl_count = ul_count = 0 ;
	peer_port = 0;
	peer_ip.s_addr   = 0;
	last_err = 0;
	type = TYPE;
	if (PORT == 0 || (type != SOCK_STREAM && type != SOCK_DGRAM)) {
		printf("ERROR %s %i\n", __FILE__, __LINE__);
		addEvent(16);
		return;
		}
	fd = socket (AF_INET, type, 0);
	if (fd == INVALID_SOCKET) {{{
		printError (GetErrno());
		addEvent(16);
		return;
		}}}
	Init();
	this_ip.s_addr = 0;
	struct sockaddr_in client_addr;
	memset(&client_addr, 0, sizeof client_addr);	// clear sin_zero
	client_addr.sin_family          = AF_INET;
	if ( strcmp(bind_ip, "0")  == 0 ) client_addr.sin_addr.s_addr     = INADDR_ANY;
		else client_addr.sin_addr.s_addr     = inet_addr(bind_ip);
	client_addr.sin_port            = htons(PORT);
#ifndef _WIN32
	int opt = 1; // sec to wait for data
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR    , &opt, sizeof(int));
#endif
	if (0 != bind(fd, (struct sockaddr*)&client_addr, sizeof(client_addr))) {
		int err = GetErrno();
		printf("cSocket::cSocket(port=%d,type=%s): Bind to socket failed.\n",PORT,GetSockTypeString(TYPE));
		printError (err);
		addEvent(16);
		return;
		} 
	memset(&client_addr, 0, sizeof client_addr);
	socklen_t len = sizeof(client_addr);
	getsockname(fd, (struct sockaddr*)&client_addr, &len);
	this_port = htons(client_addr.sin_port);
	printf("Listen on %15s:%5u (Proto: %u) ", inet_ntoa( client_addr.sin_addr), this_port, type);
	if (type == SOCK_STREAM)printf("(TCP) "); else
	if (type == SOCK_DGRAM )printf("(UDP) "); else
				printf("(Proto %u) ", type);
	if (this_port == pref.ports.server  ) printf("eDonkey-Server (should be forwarded)\n"); else
	if (this_port == pref.ports.client  ) printf("eDonkey-Client (should be forwarded)\n"); else
	if (this_port == pref.ports.emule   ) printf("eDonkey-UDP    (should be forwarded)\n"); else
	if (this_port == pref.ports.kademlia) printf("Overnet-UDP    (should be forwarded)\n"); else
	if (this_port == pref.ports.gui     ) printf("eDonkey-Gui    (this need password, should not be forwarded)\n"); else
	if (this_port == pref.ports.clc     ) printf("eDonkey-CLC    (this need password, should not be forwarded)\n"); else
	if (this_port == pref.ports.web     ) printf("eDonkey-Web    (this need password, should not be forwarded)\n"); else
	if (this_port == this_port	) printf("TODO dokument it\n");
	if (type == SOCK_STREAM) listen(fd, 4);		// if this happen we have to think about an fix :-(
	}}}
cSocket         ::~cSocket  () {{{
	if (event != NULL) delete event;
	if (timer != NULL) delete timer;
	if (fd != INVALID_SOCKET) unregisterSocket (fd);
	shutdown (fd, 2);
	closesocket(fd);
	if (send_buf != NULL) free (send_buf);
	if (read_buf != NULL) free (read_buf);
//	printf("DEL_FD %4i\n", fd);
	}}}
void     cSocket::Close     (void) {{{
	addEvent(16);	// let safely delete hdl event this
	}}}
void     cSocket::doAccept  (void) {{{
	do {
		if (fd_new != INVALID_SOCKET)  {
			printf("cSocket::doAccept: ERROR! An accepted connection was left dangling.\n");
			printf("  this_port=%u, fd_new=%u\n", this_port, fd_new);
			printf("  CLOSING SOCKET!\n");
			closesocket(fd_new);
			}
		struct sockaddr_in addr;
		socklen_t len = sizeof (addr);
		do {{{ 	fd_new = accept(fd, (struct sockaddr *)&addr, &len);
			if (fd_new != INVALID_SOCKET) break;
			last_err = GetErrno();
			if (last_err == EINTR) continue;
			if (last_err == SOCKET_RETRY) return;
			printf("accept()");
			printError(last_err);
			}}} while (true);
		if (fd_new == INVALID_SOCKET) return;
		peer_ip   = addr.sin_addr;
		peer_port = htons(addr.sin_port);
//		printf("handle new connection %15s:%5u\n", inet_ntoa(peer_ip), peer_port);
		Work();
		} while (true);				// Handle multiple connects
	}}}
void     cSocket::doSend    (void) {{{
	mask = mask & (~2);
	if (type == SOCK_DGRAM    ) return;
	if (fd   == INVALID_SOCKET) return;
	if (connected == 2) { doAccept (); return; }
	if (connected == 0) {{{ // maintain connection state
		last_conn = currentTime;
		connected = 1;
		ready_send = 1;
		socklen_t len = sizeof(last_err);
		getsockopt (fd, SOL_SOCKET, SO_ERROR, (char*)&last_err, &len);
		if (last_err == ECONNREFUSED) { addEvent(16); return; }
		if (last_err == EHOSTUNREACH) { addEvent(16); return; }
		if (last_err) printf("ERR 666 => %i\n", last_err);
		}}}
	if (send_len == 0) return;
	int done = 0;
	do {{{ 	done = send (fd, send_buf, send_len, 0);
		if (done > 0) break;
		ready_send = 0;
		last_err = GetErrno();
		if (last_err == EINTR			) continue;
		if (last_err == SOCKET_RETRY		) return;
		if (last_err == SOCKET_WOULD_BLOCK	) return;
		addEvent(16);
		if (last_err == ECONNREFUSED		) return; 
		if (last_err == EHOSTUNREACH		) return;
		if (last_err == 104			) return;
		if (last_err == 32) return;  //  EPIPE
		printf("675 err ");
		printError(last_err);		
		return;
		}}} while (true);
	last_send = currentTime;
	if (done <= 0) return;
	if (prio != 0) {{{ // Handle traffic accounting
		global_dl_count += TCP_ACK;
		//we do only count all finished bytes of that file
		//dl_count += TCP_ACK;
		global_ul_count += TCP_ACK + done;
		//ul_count += TCP_ACK + done ;
		dl_left -= TCP_ACK;
		ul_left -= TCP_ACK + done;
		}}}
	send_len -= done;
	if (send_len > 0) {
		memcpy (send_buf, send_buf + done, send_len);
		send_buf = (char*)REALLOC(send_buf, send_len);
		}
	else	{
		free (send_buf);
		send_buf = NULL;
		}
	ready_send = (send_len == 0);
	}}}
void     cSocket::doRead    (void) {{{
	if (16 == (16 & mask)) return;
	if (fd == INVALID_SOCKET) {
		addEvent(16);
		return;
		}
	int done  = 0;
	int err	  = 0;
	if (connected == 2) { doAccept (); return; }
	do {{{ 	done = ioctlsocket(fd, FIONREAD, &read_avail);
		if (done == 0) break;
		err = GetErrno();
		if (err == EINTR) continue;
		printf("ioctlsocket %15s", inet_ntoa(peer_ip));
		printError(err);
		addEvent(16);
		return;
		}}} while(true);
	if (type == SOCK_DGRAM ) {{{ // return
		if (read_avail == 0) return;
		last_read = currentTime;
		struct sockaddr_in addr;
		do {	
			ASSERT (read_buf == NULL);
			read_buf = (char*)MALLOC (read_avail);
			socklen_t s_len = sizeof(addr);
			do {{{ done = recvfrom(fd, read_buf, read_avail, 0, (struct sockaddr *)&addr, &s_len);
				if (done >= 0) break;
				err = GetErrno();
				if (err != EINTR) break;
				}}} while(true);
			if (done <= 0) {{{ // return
				free (read_buf);
				read_buf = NULL;
				return;
				}}}
			if (prio != 0) {{{ // count the traffic
				//global_dl_count += TCP_ACK + done;
				//dl_count 	+= TCP_ACK + done;
				dl_left  	-= TCP_ACK + done;
				}}}
			peer_ip   = addr.sin_addr;
			peer_port = htons(addr.sin_port);
			read_len  = done;
			doRead_high();
			if (read_buf != NULL) free (read_buf);
			read_buf    = NULL;
			read_avail -= done;
			last_read = currentTime;
			} while (read_avail > 0);
		return;
		}}}
	if (read_avail <= 0) read_avail = 4;
	if (type == SOCK_STREAM) {{{ // return
		read_buf = (char*)REALLOC (read_buf, read_len + read_avail);
		do {{{	done = recv (fd, read_buf + read_len, read_avail, 0);
			if (done == 0) {
//				printf("FIN_FD %4u - %15s\n", fd, inet_ntoa(peer_ip));
				addEvent(16);
				return;
				}
//			printf("%5i %s recv () = %i - %15s - %i\n", __LINE__, __FILE__, done, inet_ntoa(peer_ip), errno);
			int done_old = done;
			if (done > 0) break;
			err = GetErrno();
			if (err == EINTR) continue;
			done = 0;
			if (err == SOCKET_RETRY      ) break;
			if (err == SOCKET_WOULD_BLOCK) break;
			addEvent(16);
//			printf("%5i %s recv () = %i - %15s\n", __LINE__, __FILE__, done_old, inet_ntoa(peer_ip));
			if (err == ECONNREFUSED      ) return;
			if (err == 0		     ) return;
			if (err == ENOTCONN	     ) return; // 107 notconnected
			if (err == ECONNRESET	     ) return; // 104 connection reset by peer
			if (err == EHOSTUNREACH	     ) return; // 113 connection reset by peer
			if (err == ETIMEDOUT         ) return; // 110 ETIMEDOUT
			if (err == 10054	     ) return;
			if (done_old == 0 && read_avail > 0) return;
			printf("%5i %s recv () = %i\n", __LINE__, __FILE__, done_old);
			printError(err);
			return;
			}}} while (true);
		if (done > 0) {{{ // we have realy read someting
			if (connected == 0) {
				connected  = 1;
				ready_send = 1;
				last_conn = currentTime;
				}
			read_len += done;
			last_read = currentTime;
			if (prio != 0) {{{ // count traffic
				global_ul_count += TCP_ACK;
				//ul_count += TCP_ACK;
				global_dl_count += TCP_ACK + done;
				//dl_count += TCP_ACK + done ;
				ul_left -= TCP_ACK;
				dl_left -= TCP_ACK + done;
				}}}
//			if (Type == tGui) printf("cSocket::doRead %i\n", done);
			}}}
		read_buf = (char*)REALLOC (read_buf, read_len);
		return;
		}}}
	}}}
bool     cSocket::Write     (const char *buf, size_t len) {{{
	if (type == SOCK_DGRAM) {{{ // Handle UDP sending
		struct sockaddr_in addr;
		bzero (&addr, sizeof(addr));
		socklen_t addr_len = sizeof(addr);
		addr.sin_addr	= peer_ip;
		addr.sin_port   = htons(peer_port);
		addr.sin_family = AF_INET;
		last_send = currentTime;
		ASSERT(buf[0]!= 0x07);
		int ret = sendto (fd, buf, len, 0 , (struct sockaddr*)&addr, addr_len);
		if (ret < 0) {{{ // display error
			int err = GetErrno();
			if (err == 22) return false;
			if (err == 101) {
				printf("UNREACHABLE %s\n", inet_ntoa(addr.sin_addr));
				return false;
				}
			printf("IP 13 %s\n", inet_ntoa(addr.sin_addr));
			printf("LEN %i\n", len);
			if (err == 13) printf("IP 13 %s\n", inet_ntoa(addr.sin_addr));
			printError(err);
			return false;
			}}}
		if (prio != 0) {{{ //  count traffic
			global_ul_count += TCP_ACK + ret;
			//ul_count += TCP_ACK + ret ;
			ul_left -= TCP_ACK + ret;
			}}}	
		return (size_t)ret == len;
		}}}
	if (fd == INVALID_SOCKET) return false;
	send_buf = (char*)REALLOC (send_buf, send_len + len);
	memcpy (send_buf + send_len, buf, len);
	send_len += len;
	if (0 == (mask & 2)) addEvent(2);
	return (ready_send == 1);
	}}}
size_t   cSocket::Read      (      char *buf, size_t len) {{{
	if (fd   == INVALID_SOCKET) return 0;
	if (read_len < len) len = read_len;
	if (len == 0) return 0;
	ASSERT(read_len > 0    );
	ASSERT(read_buf != NULL);
	memcpy (buf, read_buf, len);
	read_len -= len;
	if (read_len == 0) {
		free (read_buf);
		read_buf = NULL;
		}
	else	{
		memcpy (read_buf, read_buf + len, read_len);
		read_buf = (char*)REALLOC(read_buf, read_len);
		}
	return len;
	}}}
bool     cSocket::doInternal(int Event) {{{
	ASSERT (fd != INVALID_SOCKET);
	if (4 == (4 & Event)) {	// Error
		socklen_t len = sizeof(last_err);
		getsockopt (fd, SOL_SOCKET, SO_ERROR, (char*)&last_err, &len);	// Error happend
//		printError(last_err);
		addEvent(16);	// error's should not be able to recover
		if (last_err == 0	     ) return false;
		if (last_err == ECONNREFUSED ) return false;	// 111
		if (last_err == ECONNRESET   ) return false;	// 104
		if (last_err == EHOSTUNREACH ) return false;	// 113
		if (last_err == 32	     ) return false; // EPIPE because read ignore read len = 0
		return false;
		}
	ASSERT (fd != INVALID_SOCKET);
	if (2 == (2 & Event)) doSend ();	// Send data if posible
	if (send_len == 0 && 32 == (mask & 32)) mask = 16;
	ASSERT (fd != INVALID_SOCKET);
	if (mask & 16) return false;
	if (1 == (1 & Event)) doRead ();
	ASSERT (fd != INVALID_SOCKET);
	if (16 == (mask & 16)) return false;
	if (fd != INVALID_SOCKET) return false;
	if (type == SOCK_DGRAM) return false;
	return (send_len == 0 || read_len > 0);
	}}}
void     cSocket::doWork    (int Event) {{{
	ASSERT (fd != INVALID_SOCKET);
	if (mask & 16) return;
	Event = Event | mask;
	mask &= 32;
	bool work = (8 == (8 & Event));
	ASSERT (fd != INVALID_SOCKET);
	if (connected == 0) work = true;
	doInternal (Event);
	if (16 == (mask & 16)) return;
	if (connected == 2) return;
	if (!work && (read_len == 0) && (send_len > 0)) return;
	bool  worked = false;
	while (read_len > 0 && fd != INVALID_SOCKET) {
		doRead_high();
		if (mask & 16) return;
		if (!canWork()) break;
		readWork();
		mask = mask & (~8);
		Work();
		work = false;
		worked = true;
		}
	if (work) {
		Work   ();
		worked = true;
		}
	if (worked) {
		work_last = currentTime;
		if (timer != NULL) delete timer;
		timer = timer_list.add_back (this);
		}
	else	{
		if (timer == NULL) timer = timer_list.add_back (this);
		}
	bool sende = (2 == (2 & mask));
	mask = mask & (~2);
	if (sende) doSend ();
	if (send_len == 0 && 32 == (mask & 32)) mask = 16;
	}}}
void     cSocket::addEvent  (int Event) {{{
	if (Event == 0) return;
	mask |= Event;
	if (event != NULL) return;
	if (prio == 0) 	event = event_list.add_front(this);
	else		event = event_list.add_back (this);
	}}}
void     cSocket::hdlEvent  (void     ) {{{
	class cSocket *stop = NULL;
	while (!event_list.empty() && ul_left > 0 && dl_left > 0) {
		class cSocket *akt = event_list.peek();
		if (stop == akt) return;
		cSocketList::remove(&akt->event);
		unsigned mask_tmp = akt->mask;
		akt->mask &= 32;
		if (!akt->validSocket()) mask_tmp = 16;
		if (mask_tmp == 0) continue;
		if (0 == (16 & mask_tmp)) akt->doWork(mask_tmp);
		if (mask_tmp & 16) delete akt;
		else if (akt->event != NULL && stop == NULL) stop = akt;
		}
	}}}

struct in_addr cSocket::My_IP (void) {{{
	if (this_ip.s_addr != 0 && !connected) return this_ip;
	struct sockaddr_in sa;
	struct in_addr null;
	null.s_addr = 0;
	socklen_t sa_len = sizeof (sa);
	if (0 != getsockname (fd, (struct sockaddr*)&sa, &sa_len)) return null;
	my_ip.s_addr = this_ip.s_addr = sa.sin_addr.s_addr;
	return sa.sin_addr;
	}}}

