#ifndef __CYGWIN_H
#define __CYGWIN_H

#include <stddef.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string>  // for stl string

#include "config.h"


#define ctol32(x)       (x)
#define ctol16(x)       (x) 

// WIN32: MSVC++ and MinGW
#if defined (_WIN32)
	// for MinGW ONLY
	#if defined (__MINGW32__)
//		#define _NO_OLDNAMES  1
//		#define NO_STRICT  1
//		#define strdup  _strdup
	#endif /* __MINGW32__ */

	// for both MSVC++ and MinGW
	typedef unsigned char      uint8_t ;
	typedef char                int8_t ;
	typedef unsigned short    uint16_t ;
	typedef short              int16_t ;
	typedef unsigned int      uint32_t ;
	typedef int                int32_t ;
	typedef unsigned __int64  uint64_t ;
	typedef __int64            int64_t ;
	#define mkdir(dir,mode) (!CreateDirectory(dir, NULL))
 	#include <windows.h>
  	#include <time.h>
  	#include <dos.h>
  	#include <ctype.h>
  	#include <io.h>
	#include "dirent_win.h"
	#include <winsock.h>
  	#define socklen_t int
  	#define getcwd(buf, size) _getcwd(buf, size)	//unused but good to know
  	#ifndef HAVE_SNPRINTF
    		#define snprintf    _snprintf
  	#endif
	//  #define vsnprintf   _vsnprintf			// unused but good to know
	#ifndef DHAVE_STRNCASECMP
		#define strncasecmp	strncmp
	#endif
	#ifndef HAVE_FTRUNCATE
		#define ftruncate(fd,len)	_chsize(fd,len)
	#endif
	#ifndef HAVE_GETPID
		#define getpid()	0
	#endif
	#define __PRETTY_FUNCTION__ "PRETTY_FUNCTION"
	#define	dir_delim	'\\'
        #define HAVE_INET_ADDR  1
        #define HAVE_GETHOSTBYNAME  1
	#define HAVE_ATEXIT  1
	#define EHOSTUNREACH            WSAEHOSTUNREACH
	#define ECONNREFUSED            WSAECONNREFUSED
	#define ECONNRESET              WSAECONNRESET
	#define ETIMEDOUT		WSAETIMEDOUT
	#define ENOTCONN		WSAENOTCONN
	// file access
	#define FD_open		_open
	#define FD_close	_close
	#define FD_fsync	_commit
	#define FD_write	_write
	#define FD_read		_read
	#define	FD_dup		_dup
	#define	FD_dup2		_dup2
	#define	FD_lseek	_lseek
	#define	FD_tell		_tell
	#define bzero(dat,len)  memset(dat,0,len)
#else	// Linux
	#define FD_open		open
	#define FD_close	close
	#define FD_fsync	fsync
	#define FD_write	write
	#define FD_read		read
	#define	FD_dup		dup
	#define	FD_dup2		dup2
	#define	FD_lseek	lseek
	#define	FD_tell		tell
	#define O_BINARY	0
        #include <sys/socket.h> //					(man socket,inet_addr)
        #include <netinet/in.h>	// Roman Shterenzon (BSD need this)	(man socket,inet_addr)
        #include <arpa/inet.h>	// by Frank Thieme			(man socket,inet_addr)
	#include <netdb.h>	// for DNS resolution			(man gethostbyname)
	#include <strings.h>
	#include <dirent.h>
	#include <unistd.h>
	#include <sys/ioctl.h>
	#include <sys/time.h>
	#include <sys/resource.h>
	#include <sys/stat.h>
	#include <signal.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <netinet/tcp.h>
	#include <sys/poll.h>
	#define dir_delim	'/'
	#define HAVE_SIGHUP	1
	#define HAVE_SIGINT	1
	#define HAVE_SIGPIPE	1
	typedef unsigned long long uint64_t;
	typedef long long int64_t;
	#define closesocket             close
	#define ioctlsocket             ioctl
#endif

#ifndef HAVE_BZERO
	#define bzero(dat,len)	memset(dat,0,len)
#endif

#ifndef __FreeBSD__
	#define HAVE_ASYNCIO		1
#else
	#define HAVE_POLLIO		1
	#define	SOL_TCP	SOL_SOCKET
#endif

/*
# if __BYTE_ORDER == __BIG_ENDIAN
#       define ctol32(x)     __bswap_32 (x)
#       define ctol16(x)     __bswap_16 (x)
#endif
*/
#if __BYTE_ORDER == __LITTLE_ENDIAN
#       define ctol32(x)       (x)
#       define ctol16(x)       (x)
#endif
	
#define ASSERT(x)	{ bool assert_test = (x); if (!assert_test) { fflush(stdout); assert(x); } }
inline  void* MALLOC (size_t x) { void *ret = malloc(x); assert(ret != NULL); return ret; }
inline  char* STRDUP (const char *x) { char *ret = strdup(x); assert(ret != NULL); return ret; }
inline  void* REALLOC(void *ptr, size_t size) { void *ret = realloc(ptr, size); if (size != 0) assert(ret != NULL); return ret; }

#endif // __CYGWIN_H
